from typing import Literal, Union

import numpy as np
from scipy.spatial.transform.rotation import Rotation

from alitra.frame_dataclasses import (
    PointList,
    Point,
    Transform,
    Quaternion,
)


class FrameTransform:
    """Let (from_) be the fixed local coordinate frame that the robot operate in, and (to_) be
    in the asset-fixed global coordinate system. Further, let the relationship between the two
    reference systems be described by : p_to = rotation_object.apply(p_from) + translation.
    This object allows for transform between the frames (from_) and (to_). Can be rewritten to
    accommodate for multiple frames.
    :param transform: A transform dataclass object
    """

    def __init__(self, transform: Transform):
        self.transform = transform

    def transform_point(
        self,
        coordinates: Union[Point, PointList],
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Union[Point, PointList]:
        """
        Transforms a point or list of points from _from to _to (rotation and translation)
        :param coordinates: Point or PointList of coordinates to transform.
        :param from_: Source coordinate system. Must be "robot" or "asset" and different to "to_".
        :param to_: Destination coordinate system. Must be "robot" or "asset" and different to "from_".
        :return: Point or PointList with coordinates in the to_ coordinate system.
        """
        if coordinates.frame != from_:
            raise ValueError(
                f"Expected coordinates in frame {from_} "
                + f", got coordinates in frame {coordinates.frame}"
            )

        result: np.ndarray
        if from_ == self.transform.to_ and to_ == self.transform.from_:
            """Using the inverse transform"""
            result = self.transform.rotation_object.apply(
                coordinates.as_np_array() - self.transform.translation.as_np_array(),
                inverse=True,
            )

        elif from_ == self.transform.from_ and to_ == self.transform.to_:
            result = (
                self.transform.rotation_object.apply(coordinates.as_np_array())
                + self.transform.translation.as_np_array()
            )
        else:
            raise ValueError("Transform not specified")

        if isinstance(coordinates, Point):
            return Point.from_array(result, to_)
        elif isinstance(coordinates, PointList):
            return PointList.from_array(result, to_)
        else:
            raise ValueError("Incorrect input format. Must be Point or PointList.")

    def transform_quaternion(
        self,
        quaternion: Quaternion,
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Quaternion:
        quaternion_from: Rotation = Rotation.from_quat(quaternion.as_np_array())

        if not isinstance(quaternion, Quaternion):
            raise ValueError("Incorrect input format. Must be Quaternion.")

        if from_ == self.transform.to_ and to_ == self.transform.from_:
            """Using the inverse transform"""

            quaternion_to = quaternion_from * self.transform.rotation_object.inv()
        elif from_ == self.transform.from_ and to_ == self.transform.to_:
            quaternion_to = quaternion_from * self.transform.rotation_object
        else:
            raise ValueError("Transform not specified")

        result: Quaternion = Quaternion.from_array(quaternion_to.as_quat(), frame=to_)
        return result



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal

import numpy as np
from scipy.spatial.transform import Rotation


@dataclass
class Transform:
    """
    A transform object that describe the transformation between two frames.
    Euler or quaternion must be provided to perform a rotation. If no rotation is required specify the unit quaternion
    or zero Euler angles. Translations must be expressed in the (to_) frame
    """

    translation: Translation
    from_: Literal["robot", "asset"]
    to_: Literal["asset", "robot"]
    rotation_object: Rotation = None

    def __post_init__(self):
        if (
            not self.translation.from_ == self.from_
            and not self.translation.to_ == self.to_
            or not self.to_ == self.translation.frame
        ):
            raise ValueError(
                f"The from_ frames or to_ frames of translation and transform object are not equal."
            )

    @staticmethod
    def from_euler_ZYX(
        translation: Translation,
        from_: Literal["robot", "asset"],
        to_: Literal["asset", "robot"],
        euler: Euler,
    ):
        rotation_object = Rotation.from_euler("ZYX", euler.as_np_array())
        return Transform(
            translation=translation,
            from_=from_,
            to_=to_,
            rotation_object=rotation_object,
        )

    @staticmethod
    def from_quat(
        translation: Translation,
        from_: Literal["robot", "asset"],
        to_: Literal["asset", "robot"],
        quat: Quaternion,
    ):
        rotation_object = Rotation.from_quat(quat.as_np_array())
        return Transform(
            translation=translation,
            from_=from_,
            to_=to_,
            rotation_object=rotation_object,
        )


@dataclass
class Point:
    x: float
    y: float
    frame: Literal["robot", "asset"]
    z: float = 0

    def as_np_array(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z], dtype=float)

    @staticmethod
    def from_array(coordinate: np.ndarray, frame: Literal["robot", "asset"]) -> Point:
        if coordinate.shape != (3,):
            raise ValueError("Point should have shape (3,)")
        return Point(x=coordinate[0], y=coordinate[1], z=coordinate[2], frame=frame)


@dataclass
class PointList:
    points: List[Point]
    frame: Literal["robot", "asset"]

    def as_np_array(self) -> np.ndarray:
        points = []
        for point in self.points:
            points.append([point.x, point.y, point.z])
        return np.array(points, dtype=float)

    @staticmethod
    def from_array(
        point_array: np.ndarray, frame: Literal["robot", "asset"]
    ) -> PointList:
        if point_array.shape[1] != 3:
            raise ValueError("Coordinate_list should have shape (3,N)")
        points: List[Point] = []
        for point in point_array:
            points.append(Point(x=point[0], y=point[1], z=point[2], frame=frame))
        return PointList(points=points, frame=frame)


@dataclass
class Euler:
    """Euler angles where (psi,theta,phi) is rotation about the z-,y-, and x- axis respectively as_array and from_array
    are both ordered by rotation about z,y,x. That is [psi,theta,phi]"""

    from_: Literal["robot", "asset"]
    to_: Literal["robot", "asset"]
    psi: float = 0
    theta: float = 0
    phi: float = 0

    def as_np_array(self) -> np.ndarray:
        return np.array([self.psi, self.theta, self.phi], dtype=float)

    @staticmethod
    def from_array(
        rotations: np.ndarray,
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Euler:
        if rotations.shape != (3,):
            raise ValueError("Coordinate_list should have shape (3,)")
        return Euler(
            psi=rotations[0], theta=rotations[1], phi=rotations[2], from_=from_, to_=to_
        )


@dataclass
class Quaternion:
    frame: Literal["robot", "asset"]
    x: float = 0
    y: float = 0
    z: float = 0
    w: float = 1

    def as_np_array(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z, self.w], dtype=float)

    @staticmethod
    def from_array(
        quaternion: np.ndarray,
        frame: Literal["robot", "asset"],
    ) -> Quaternion:
        if quaternion.shape != (4,):
            raise ValueError("Quaternion array must have shape (4,)")
        return Quaternion(
            x=quaternion[0],
            y=quaternion[1],
            z=quaternion[2],
            w=quaternion[3],
            frame=frame,
        )


@dataclass
class Translation:
    """Translations should be expressed in the to_ frame, which are typically the asset frame"""

    x: float
    y: float
    from_: Literal["robot", "asset"]
    to_: Literal["robot", "asset"]
    frame: Literal["asset", "robot"] = "asset"
    z: float = 0

    def as_np_array(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z], dtype=float)

    @staticmethod
    def from_array(
        coordinate: np.ndarray,
        from_: Literal["robot", "asset"],
        to_: Literal["robot", "asset"],
    ) -> Translation:
        if coordinate.shape != (3,):
            raise ValueError("Point should have shape (3,)")
        return Translation(
            x=coordinate[0], y=coordinate[1], z=coordinate[2], from_=from_, to_=to_
        )



from scipy.spatial.transform.rotation import Rotation

from alitra import Euler, Quaternion


def quaternion_to_euler(
    quaternion: Quaternion, sequence: str = "ZYX", degrees: bool = False
) -> Euler:
    """
    Transform a quaternion into Euler angles.
    :param quaternion: A Quaternion object.
    :param sequence: Rotation sequence for the Euler angles.
    :param degrees: Set to true if the resulting Euler angles should be in degrees. Default is radians.
    :return: Euler object.
    """
    rotation_object: Rotation = Rotation(quaternion.as_np_array())
    euler: Euler = Euler.from_array(
        rotation_object.as_euler(sequence, degrees=degrees),
        from_="robot",
        to_="asset",
    )
    return euler



"""
Alitra makes alignment and transformation between coordinate-frames
easier by using custom dataclasses

Imagine we have an "asset" with one coordinate-frame and a "robot" with its own
internal coordinate-frame. We want to transform a point on the robot to a point in the
asset coordinate-frame.

>>> import numpy as np
>>> from alitra.align_frames import AlignFrames
>>> from alitra.frame_dataclasses import Euler, Quaternion, PointList, Translation
>>> from alitra.frame_transform import FrameTransform

Setting up rotations, translations and points for transformation

>>> eul_rot = Euler(psi=np.pi / 4, from_="robot", to_="asset").as_np_array()
>>> ref_translations = Translation(x=1, y=0, from_="robot", to_="asset")
>>> p_robot = PointList.from_array(np.array([[1, 1, 0], [10, 1, 0]]), frame="robot")
>>> rotation_axes = "z"

Making the transform

>>> c_frame_transform = FrameTransform(
...    eul_rot, ref_translations, from_=eul_rot.from_, to_=eul_rot.to_
... )

Tranform point on robot to a point on the asset

>>> p_asset = c_frame_transform.transform_point(p_robot, from_="robot", to_="asset")

If you have one point in two different frames and the rotation between the axes you can find the transform
between the two frames.

>>> transform = AlignFrames.align_frames(p_robot, p_asset, rotation_axes)

"""

from alitra.align_frames import AlignFrames
from alitra.frame_dataclasses import (
    Euler,
    Quaternion,
    Point,
    PointList,
    Transform,
    Translation,
)
from alitra.frame_transform import FrameTransform
